#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>
#include <SDL.h>
#include <SDL_image.h>
#include <SDL_gfxPrimitives.h>
#include <SDL_rotozoom.h>
#include <sys/time.h>
#include <getopt.h>
#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>
#include "xyssl/md5.h"
#include "icon.xpm"
#include "fill.h"

/* OS specific includes */
#ifdef _WIN32
#define _LONG "%I64d"
#include <windows.h>
#else
#define _LONG "%lld"
#include <sys/resource.h>
#ifndef _PSP_FW_VERSION
#include <sys/utsname.h>
#endif
#endif

/* useful constants */
#define VERSION "3.1"
#define TIMER 60 * 60 * 1000
#ifdef _PSP_FW_VERSION
#define X 272
#define Y 272
#else
#define X 608
#define Y 608
#endif
#define W X/WIDTH
#define H Y/HEIGHT

/* useful macros */
#ifdef _PSP_FW_VERSION
#define LOG(x, y...) do { pspDebugScreenPrintf("%s ", timestamp()); pspDebugScreenPrintf(y); } while(0)
#else
#define LOG(x, y...) do { fprintf(x, "%s ", timestamp()); fprintf(x, y); fflush(x); } while(0)
#endif

/* global variables */
piece pieces[WIDTH * HEIGHT + 1];
piece rotations[4 + 4 * WIDTH * HEIGHT];
int curgrid[WIDTH][HEIGHT], bestgrid[WIDTH][HEIGHT], ishint[WIDTH][HEIGHT];
char onboard[WIDTH * HEIGHT + 1];
int prio[WIDTH * HEIGHT + 1];
int prio_nb = 0;
int havecolors[COLORS][COLORS][4 + 4 * WIDTH * HEIGHT + 1];
int best = 0, seed, solutions = 0;
long long pps = 0, pps_optim = 0, pps_swap = 0, pps_total = 0, pps_restart;
SDL_Surface *screen;
int x[WIDTH*HEIGHT], y[WIDTH*HEIGHT];
float coeff1 = 192.0, coeff2 = 11.0, coeff3 = 1.8, coeff4 = 10.0, coeff5 = 10.0, coeff6 = 0, coeff7 = 2, coeff8 = 99;
int maxskip[WIDTH * HEIGHT];
long long lastsync;

/* data sent for stats */
char username[SIZE];
char password[SIZE];
char hostname[SIZE];
char os[SIZE];
char release[SIZE];
char machine[SIZE];

/* user interface */
Uint32 palette[COLORS];
SDL_Surface *tiles[COLORS];
int gui = 0;
int minigui = 0;
int screensaver = 0;
int benchmark = 0;
int drawnow = 0;
int drawnow_mini = 0;
int startup = 1;

/* timestamp */
char _timestamp[30];
char *timestamp()
{
	time_t t;
	struct tm *lt;
	t = time(NULL);
	lt = localtime(&t);
	sprintf(_timestamp, "%.4d-%.2d-%.2d %.2d:%.2d:%.2d #%.5d |", lt->tm_year+1900, lt->tm_mon+1, lt->tm_mday, lt->tm_hour, lt->tm_min, lt->tm_sec, getpid());
	return _timestamp;
}

void error(char *message)
{
	LOG(stderr, "Error: %s!\n", message);
	#ifdef _PSP_FW_VERSION
	sceKernelSleepThread();
	#endif
	exit(1);
}

void warning(char *message)
{
	LOG(stderr, "Warning: %s!\n", message);
}

Uint32 draw_timer()
{
	drawnow = 1;
	return 50;
}

Uint32 drawmini_timer()
{
	drawnow_mini = 1;
	return 1000;
}

Uint32 benchmark_timer()
{
	LOG(stderr, "Stats: pps = " _LONG "\n", pps);
	pps = 0;
	return 1000;
}

#ifdef DEBUG
Uint32 debug_timer()
{
	LOG(stderr, "Stats: best = %d, c1 = %.2f, c2 = %.2f, c3 = %.2f, c4 = %.2f, c5 = %.2f, c6 = %f, c7 = %f, c8 = %f, pps = %8lld, pps_optim = %8lld, pps_swap = %8lld, %6.2f %%\n", best, coeff1, coeff2, coeff3, coeff4, coeff5, coeff6, coeff7, coeff8, pps, pps_optim, pps_swap, 100.0*(pps_optim+pps_swap)/pps);
	pps = 0;
	pps_optim = 0;
	pps_swap = 0;
	return 1000;
}
#endif

size_t curl_write(void *ptr, size_t size, size_t nb, void *str)
{
	int t = nb * size;
	if (t > SIZE - 1) t = SIZE - 1;
	memcpy(str, ptr, t);
	return nb * size;
}

Uint32 stats_timer()
{
	int i, j, tim, key;
	char request[SIZE*10], response[SIZE], temp[SIZE];
	CURL *curl;
	struct timeval tv;
	long long newsync;
	
	gettimeofday(&tv, NULL);
	newsync = (long long) tv.tv_sec * 1000000 + tv.tv_usec;
	if (startup) lastsync = newsync;
	
	tim = time(NULL);
	key = tim ^ pps ^ 0x12345678;
	
	/* the big HTTP request */
	sprintf(request,
		"version=" VERSION
		"&timer=" _LONG
		"&username=%s"
		"&password=%s"
		"&hostname=%s"
		"&os=%s"
		"&release=%s"
		"&machine=%s"
		"&seed=%d"
		"&pps=" _LONG
		"&pps_optim=" _LONG
		"&pps_swap=" _LONG
		"&coeff1=%f"
		"&coeff2=%f"
		"&coeff3=%f"
		"&coeff4=%f"
		"&coeff5=%f"
		"&coeff6=%f"
		"&coeff7=%f"
		"&coeff8=%f"
		"&best=%d"
		"&time=%d"
		"&key=%d",
		(newsync - lastsync) / 1000, username, password, hostname, os, release, machine, seed, pps, pps_optim, pps_swap, coeff1, coeff2, coeff3, coeff4, coeff5, coeff6, coeff7, coeff8, best, tim, key
	);
	
	lastsync = newsync;
	
	/* add the best grid */
	strcat(request, "&grid=");
	for (j = 0; j < HEIGHT; j++) for (i = 0; i < WIDTH; i++)
	{
		sprintf(temp, "%d,", bestgrid[i][j]);
		strcat(request, temp);
	}
	
	/* add the hints */
	strcat(request, "&hints=");
	for (j = 0; j < HEIGHT; j++) for (i = 0; i < WIDTH; i++) if (ishint[i][j] == 3)
	{
		sprintf(temp, "%d:%dx%d:%d,", curgrid[i][j]/4, i+1, j+1, (4-curgrid[i][j]%4)%4);
		strcat(request, temp);
	}
	
	strcat(request, "&junk=\n");
	
	/* default response code */
	memset(response, 0, SIZE);
	if (pps == 0)
		response[0] = 'e';
	else
		response[0] = 'w';
	
	/* post the data */
	curl = curl_easy_init();
	curl_easy_setopt(curl, CURLOPT_USERAGENT, "Eternity2 " VERSION);
	curl_easy_setopt(curl, CURLOPT_TIMEOUT, 30);
	curl_easy_setopt(curl, CURLOPT_NOSIGNAL, 1);
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, curl_write);
	curl_easy_setopt(curl, CURLOPT_WRITEDATA, &response);
	curl_easy_setopt(curl, CURLOPT_POSTFIELDS, request);
	curl_easy_setopt(curl, CURLOPT_URL, "http://www.eternity2.fr/e2/timer.php");
	curl_easy_perform(curl);
	curl_easy_cleanup(curl);
	
	/* handle response code */
	switch (response[0])
	{
		case 'o':
			/* try to setup coeffs if available */
			#ifndef DEBUG
			sscanf(response, "o coeff1=%f coeff2=%f coeff3=%f coeff4=%f coeff5=%f coeff6=%f coeff7=%f coeff8=%f prio=%s", &coeff1, &coeff2, &coeff3, &coeff4, &coeff5, &coeff6, &coeff7, &coeff8, temp);
			#endif
			for (i = 0; i < WIDTH * HEIGHT; i++)
				maxskip[i] = i < coeff1 ? -1 : pow( (i - coeff1) / coeff2, coeff3);
			/* load priority table */
			prio_nb = 0;
			for (i = 0; temp[i*2] != '.'; i++)
			{
				char hex[3];
				int code;
				strncpy(hex, &temp[i*2], 2);
				hex[2] = 0;
				sscanf(hex, "%X", &code);
				prio[i] = code ^ 66;
				prio_nb++;
			}
			LOG(stdout, "Synchronized with server. Best score is %d.\n", best);
			break;
		case 'e':
			error("unable to contact server, please check your internet connection");
			break;
		case 'w':
			warning("unable to contact server, recent stats will be lost");
			break;
		case 'v':
			error("a new version is available, please upgrade the software");
			break;
		case 'k':
			error("invalid key");
			break;
		case 'i':
			error("invalid username/password");
			break;
		default:
			if (startup)
				error("unknown response from server");
			/* quick fix for proxies */
			else
				warning("unknown response from server");
			break;
	}
	
	/* reset pps */
	pps = 0;
	pps_optim = 0;
	pps_swap = 0;
	
	/* go again in 1 second */
	return TIMER;
}

void quit()
{
	/* try to send the stats before shutdown */
	if (benchmark)
		LOG(stdout, "Final stats: solutions = %d, total = " _LONG "\n", solutions, pps_total);
	else
		stats_timer();
	SDL_Quit();
	
	/* end message */
	LOG(stdout, "Ending...\n");
	
	exit(0);
}

void draw_mini()
{
	char temp[20];
	SDL_Event event;
	
	sprintf(temp, "Best = %d", best);
	SDL_FillRect(screen, NULL, SDL_MapRGB(screen->format, 0, 0, 0));
	stringRGBA(screen, 20, 10, temp, 255, 255, 255, 255);
	SDL_Flip(screen);
	
	while (SDL_PollEvent(&event)) switch (event.type)
	{
		case SDL_QUIT:
			quit();
		case SDL_KEYDOWN:
		case SDL_MOUSEBUTTONDOWN:
			if (screensaver) quit();
	}
	
	drawnow_mini = 0;
}

void draw()
{
	int i, j;
	SDL_Event event;
	SDL_Rect src, dst;
	
	SDL_FillRect(screen, NULL, palette[0]);
	for (j = 0; j < HEIGHT; j++) for (i = 0; i < WIDTH; i++) if (curgrid[i][j]/4 > 0)
	{
		int xa, xb, xc, xd, xe;
		int ya, yb, yc, yd, ye;
		
		xa = xc = i * W;
		xb = xd = i * W + W;
		xe = i * W + W / 2;
		
		ya = yb = j * H;
		yc = yd = j * H + H;
		ye = j * H + H / 2;
		
		src.x = 0; src.y = H/2; src.w = W; src.h = H/2;
		dst.x = xa; dst.y = ya;
		SDL_BlitSurface(tiles[(int)rotations[curgrid[i][j]][up]], &src, screen, &dst);
		src.x = 0; src.y = 0; src.w = W; src.h = H/2;
		dst.x = xa; dst.y = ye;
		SDL_BlitSurface(tiles[(int)rotations[curgrid[i][j]][down]], &src, screen, &dst);
		src.x = W/2; src.y =0; src.w = W/2; src.h = H;
		dst.x = xa; dst.y = ya;
		SDL_BlitSurface(tiles[(int)rotations[curgrid[i][j]][left]], &src, screen, &dst);
		src.x = 0; src.y =0; src.w = W/2; src.h = H;
		dst.x = xe; dst.y = ya;
		SDL_BlitSurface(tiles[(int)rotations[curgrid[i][j]][right]], &src, screen, &dst);
		
		lineRGBA(screen, xa, ya, xb, yb, 0, 0, 0, 255);
		lineRGBA(screen, xb, yb, xd, yd, 0, 0, 0, 255);
		lineRGBA(screen, xc, yc, xd, yd, 0, 0, 0, 255);
		lineRGBA(screen, xa, ya, xc, yc, 0, 0, 0, 255);
		lineRGBA(screen, xa, ya, xd, yd, 0, 0, 0, 255);
		lineRGBA(screen, xb, yb, xc, yc, 0, 0, 0, 255);
	}
	SDL_Flip(screen);
	
	while (SDL_PollEvent(&event)) switch (event.type)
	{
		case SDL_QUIT:
			quit();
		case SDL_KEYDOWN:
		case SDL_MOUSEBUTTONDOWN:
			if (screensaver) quit();
	}
	
	drawnow = 0;
}

/* display the score in the window title */
void update_caption()
{
	char temp[40];
	sprintf(temp, "Best: %d | Eternity2.fr " VERSION, best);
	SDL_WM_SetCaption(temp, temp);
}

void save(int grid[WIDTH][HEIGHT], int score)
{
	int i, j;
	
	memcpy(bestgrid, grid, sizeof(bestgrid));
	best = score;
	if (benchmark && score == WIDTH * HEIGHT * 2 - WIDTH - HEIGHT)
	{
		solutions++;
		LOG(stdout, "Solution:\n");
		for (j = 0; j < HEIGHT; j++)
		{
			LOG(stdout, "     ");
			for (i = 0; i < WIDTH; i++) fprintf(stdout, "%3d ", grid[i][j]/4);
			fprintf(stdout, "\n");
			fflush(stdout);
		}
	}
	#if DEBUG > 1
	LOG(stdout, "Score %d:\n", score);
	for (j = 0; j < HEIGHT; j++)
	{
		LOG(stdout, "     ");
		for (i = 0; i < WIDTH; i++) fprintf(stdout, "%3d ", grid[i][j]/4);
		fprintf(stdout, "\n");
		fflush(stdout);
	}
	LOG(stdout, "\n");
	#endif
	if (benchmark > 1 && score >= benchmark)
	{
		LOG(stdout, "Found %d\n", score);
		quit();
	}
	if (gui || minigui) update_caption();
}

int crc16(unsigned char *ptr, int count)
{
	int crc, i;
	
	crc = 0;
	while(--count >= 0)
	{
		crc = crc ^ (int)*ptr++ << 8;
		for(i = 0; i < 8; ++i)
			if(crc & 0x8000)
				crc = crc << 1 ^ 0x1021;
			else
				crc = crc << 1;
	}
	return (crc & 0xFFFF);
}

void shuffle()
{
	int i, j, k, l;
	
	/* list of pieces having 2 colors */
	for (i = 0; i < COLORS; i++) for (j = 0; j < COLORS; j++)
	{
		l = 0;
		for (k = 4; k < 4 + 4 * WIDTH * HEIGHT; k++)
			if (
				( (i == ANY && rotations[k][up] != 0) || rotations[k][up] == i)
				&&
				( (j == ANY && rotations[k][left] != 0) || rotations[k][left] == j)
			) havecolors[i][j][l++] = k;
		havecolors[i][j][l] = 0;
		/* shuffle the pieces order in the list */
		if (!benchmark) for (k = 0; k < l; k++)
		{
			int r = rand() % l;
			int t = havecolors[i][j][k];
			havecolors[i][j][k] = havecolors[i][j][r];
			havecolors[i][j][r] = t;
		}
		/* now place the pieces with higher priority at the top the list */
		int d = 0;
		for (k = 0; k < prio_nb; k++)
		{
			int s;
			for (s = 0; s < l; s++)
			if (havecolors[i][j][s]/4 == prio[k])
			{
				int t = havecolors[i][j][s];
				havecolors[i][j][s] = havecolors[i][j][d];
				havecolors[i][j][d] = t;
				d++;
			}
		}
	}
	
	/* reset restart counter */
	pps_restart = 0;
}

void init()
{
	int i, j, k, l;
	FILE *fic;
	struct timeval tv;
	char *nl;
	unsigned char buffer[SIZE];
	
	/* start message */
	LOG(stdout, "Starting...\n");
	
	/* we want some random values */
	gettimeofday(&tv, NULL);
	/* this is what nmap uses... */
	seed = tv.tv_sec ^ tv.tv_usec ^ getpid();
	/* this looks less linear... */
	seed = (tv.tv_sec+1) * (tv.tv_usec+1) * (getpid()+1);
	srand(seed);
	#ifdef DEBUG
	srand(1);
	#endif
	
	/* set this process to low-priority */
	#ifdef _WIN32
	SetPriorityClass(GetCurrentProcess(), IDLE_PRIORITY_CLASS);
	#else
	#ifndef _PSP_FW_VERSION
	setpriority(PRIO_PROCESS, getpid(), 19);
	#endif
	#endif
	
	/* load OS data */
	{
	#ifdef _WIN32
		OSVERSIONINFO osver;
		SYSTEM_INFO sysinfo;
		DWORD size = SIZE;
		
		memset(hostname, 0, sizeof(hostname));
		GetComputerName(hostname, &size);
		osver.dwOSVersionInfoSize = sizeof(osver);
		GetVersionEx(&osver);
		strcpy(os, "Windows");
		sprintf(release, "%d.%d", (int)osver.dwMajorVersion, (int)osver.dwMinorVersion);
		GetSystemInfo(&sysinfo);
		switch(sysinfo.wProcessorArchitecture)
		{
			case PROCESSOR_ARCHITECTURE_AMD64:
				strcpy(machine, "x64");
				break;
			case PROCESSOR_ARCHITECTURE_IA64:
				strcpy(machine, "ia64");
				break;
			case PROCESSOR_ARCHITECTURE_INTEL:
				strcpy(machine, "x86");
				break;
			default:
				strcpy(machine, "unknown");
				break;
		}
	#else
	#ifdef _PSP_FW_VERSION
		strcpy(hostname, "unknown");
		strcpy(os, "PSP");
		strcpy(release, "1.5");
		strcpy(machine, "psp");
	#else
		struct utsname u;
		uname(&u);

		strcpy(hostname, u.nodename);
		strcpy(os, u.sysname);
		strcpy(release, u.release);
		strcpy(machine, u.machine);
	#endif
	#endif
	}
	
	/* reset tables */
	memset(curgrid, 0, sizeof(curgrid));
	memset(onboard, 0, sizeof(onboard));
	memset(ishint, 0, sizeof(ishint));
	
	/* load user.txt */
	if (!benchmark)
	{
		if ((fic = fopen("user.txt", "r")) == NULL)
			error("you must create a user.txt file");
		if (fgets(username, SIZE, fic) == NULL)
			error("unable to read username in user.txt");
		if (fgets(password, SIZE, fic) == NULL)
			error("unable to read password in user.txt");
		fclose(fic);

		/* fix username */
		if ((nl = strchr(username, '\r')) != NULL) *nl = '\0';
		if ((nl = strchr(username, '\n')) != NULL) *nl = '\0';
		
		/* compute password md5 */
		if ((nl = strchr(password, '\r')) != NULL) *nl = '\0';
		if ((nl = strchr(password, '\n')) != NULL) *nl = '\0';
		md5((unsigned char*) password, strlen(password), buffer);
		for (i = 0; i < 16; i++) sprintf(password + i*2, "%.2x", buffer[i]);
	}
	
	/* load pieces.txt */
	pieces[0][up] = pieces[0][right] = pieces[0][down] = pieces[0][left] = ANY; /* not used */
	if ((fic = fopen("pieces.txt", "r")) == NULL)
		error("you must create a pieces.txt file");
	for (i = 1; i < WIDTH * HEIGHT + 1; i++) for (j = 0; j < 4; j++)
	{
		fscanf(fic, "%d", &k);
		pieces[i][j] = k;
	}
	fclose(fic);
	
	if (!benchmark)
	{
		/* check the pieces are valid */
		if (crc16((unsigned char *)&pieces[1], WIDTH * HEIGHT * sizeof(piece)) != 33591)
			error("the pieces.txt file has got an invalid CRC16, please check your pieces");
		
		/* place the hint */
		onboard[139] = 2;
		ishint[7][8] = 2;
		curgrid[7][8] = 139 * 4 + 2;
		
		/* load hints.txt */
		if ((fic = fopen("hints.txt", "r")) != NULL)
		while (fscanf(fic, "%d %d %d %d", &k, &i, &j, &l) == 4)
		{
			i--;
			j--;
			if (i < 0 || i >= WIDTH || j < 0 || j >= HEIGHT) error("invalid position for hint");
			if (k < 1 || k > 256) error("invalid piece for hint");
			if (l < 0 || l > 3) error("invalid rotation for hint");
			if (ishint[i][j]) error("there is already a hint at this position");
			if (onboard[k]) error("this hint is already on grid");
			onboard[k] = 3;
			ishint[i][j] = 3;
			curgrid[i][j] = k * 4 + (4-l)%4;
			LOG(stdout, "Hint: added piece %d at position %c%d with rotation %d.\n", k, j+'A', i+1, l);
		}
	}
	
	/* pre-compute the rotations */
	for (i = 0; i < WIDTH * HEIGHT + 1; i++) for (j = 0; j < 4; j++)
	{
		rotations[i*4+j][left]  = pieces[i][(j + left) % 4];
		rotations[i*4+j][up]    = pieces[i][(j + up) % 4];
		rotations[i*4+j][right] = pieces[i][(j + right) % 4];
		rotations[i*4+j][down]  = pieces[i][(j + down) % 4];
	}
	
	/* build the scan order */
	/* row scan */
	l = 0;
	for (j = 0; j < HEIGHT; j++) for (i = 0; i < WIDTH; i++)
	{
		x[l] = i;
		y[l] = j;
		l++;
	}
	/* corner first */
	/*
	l = 0;
	for (k = 0; k < WIDTH; k++)
	{
		for (i = 0; i < k; i++)
		{
			x[l] = i;
			y[l] = k;
			l++;
		}
		for (j = 0; j <= k; j++)
		{
			x[l] = k;
			y[l] = j;
			l++;
		}
	}
	*/
	/* top left 1/4, then top right 1/4, then bottom 1/2 */
	/*
	l = 0;
	for (j = 0; j <= HEIGHT / 2; j++)
	for (i = 0; i < WIDTH / 2; i++)
	{
		x[l] = i;
		y[l] = j;
		l++;
	}
	for (j = 0; j <= HEIGHT / 2; j++)
	for (i = WIDTH / 2; i < WIDTH; i++)
	{
		x[l] = i;
		y[l] = j;
		l++;
	}
	for (j = HEIGHT / 2 + 1; j < HEIGHT; j++)
	for (i = 0; i < WIDTH; i++)
	{
		x[l] = i;
		y[l] = j;
		l++;
	}
	*/
	
	/* read initial layout from stdin... */
	/*
	for (j = 0; j < HEIGHT; j++) for (i = 0; i < WIDTH; i++)
	{
		scanf("%d,", &k);
		if (k == 0) { l = j * WIDTH + i; i = j = 999; break; }
		curgrid[i][j] = k;
		onboard[k/4] = 1;
	}
	LOG(stderr, "Read: %d\n", l);
	*/
	
	/* capture the signals */
	signal(SIGINT, quit);
	signal(SIGTERM, quit);
	
	/* setup SDL and timer */
	i = SDL_INIT_TIMER;
	if (gui || minigui) i |= SDL_INIT_VIDEO;
	if (SDL_Init(i) == -1)
		error("SDL_Init");
	
	if (gui || minigui)
	{
		update_caption();
		SDL_WM_SetIcon(IMG_ReadXPMFromArray(icon), NULL);
		
		/* initialize the SDL window */
		i = SDL_HWSURFACE | SDL_ANYFORMAT | SDL_DOUBLEBUF;
		if (screensaver) i |= SDL_FULLSCREEN;
		screen = SDL_SetVideoMode(minigui?120:X, minigui?28:Y, 32, i);
		
		if (screen == NULL) error("SDL_SetVideoMode");
		
		if (screensaver) SDL_ShowCursor(SDL_DISABLE);
	}
	
	if (gui)
	{
		/* initialize palette */
		palette[0] = SDL_MapRGB(screen->format, 0, 0, 0);
		for (i = 1; i < COLORS; i++) palette[i] = rand()%256<<24 | rand()%256<<16 | rand()%256<<8 | 255;
		
		/* load tiles */
		for (i = 0; i < COLORS; i++)
		{
			char temp[20];
			SDL_Surface *t;
			sprintf(temp, "tiles/%d.gif", i);
			if ((t = IMG_Load(temp)) != NULL)
			{
				tiles[i] = zoomSurface(t, 1.0*W/t->w, 1.0*H/t->h, 1);
				SDL_FreeSurface(t);
				filledTrigonColor(tiles[i], 0, 0, W/2, 0, 0, H/2, 0xff);
				filledTrigonColor(tiles[i], W/2, 0, W, 0, W, H/2, 0xff);
				filledTrigonColor(tiles[i], 0, H/2, 0, H, W/2, H, 0xff);
				filledTrigonColor(tiles[i], W/2, H, W, H/2, W, H, 0xff);
				/* fix right border */
				boxColor(tiles[i], W-1, 0, W, H, 0xff);
			}
			else
			{
				t = SDL_CreateRGBSurface(SDL_SWSURFACE, W, H, 32, 0x000000ff, 0x0000ff00, 0x00ff0000, 0xff000000);
				tiles[i] = SDL_DisplayFormatAlpha(t);
				SDL_FreeSurface(t);
				SDL_FillRect(tiles[i], NULL, 0);
				filledTrigonColor(tiles[i], W/2, 0, 0, H/2, W, H/2, palette[i]);
				filledTrigonColor(tiles[i], W/2, H, 0, H/2, W, H/2, palette[i]);
			}
		}
		
		#ifndef _PSP_FW_VERSION
		SDL_AddTimer(1, draw_timer, 0);
		#endif
	}
	
	if (minigui)
		SDL_AddTimer(1, drawmini_timer, 0);
	
	#ifdef DEBUG
	SDL_AddTimer(1000, debug_timer, 0);
	#endif
	
	#ifndef _PSP_FW_VERSION
	if (benchmark)
		SDL_AddTimer(1000, benchmark_timer, 0);
	else
	{
		stats_timer();
		SDL_AddTimer(TIMER, stats_timer, 0);
	}
	#endif
	startup = 0;
}

int main(int argc, char *argv[])
{
	#ifdef _PSP_FW_VERSION
	gui = 1;
	benchmark = 0;
	screensaver = 1;
	pspDebugScreenInit();
	#else
	/* parse command */
	int c, option_index = 0;
	static char const short_options[] = "";
	static struct option long_options[] = {
		{ "gui", no_argument, NULL, 'g' },
		{ "minigui", no_argument, NULL, 'm' },
		{ "benchmark", optional_argument, NULL, 'b' },
		{ "help", no_argument, NULL, 'h' },
		{ NULL, 0, NULL, 0 }
	};
	while((c = getopt_long(argc, argv, short_options, long_options, &option_index)) != -1)
		switch (c)
		{
			case 'g':
				gui = 1;
				break;
			case 'm':
				minigui = 1;
				break;
			case 'b':
				if (optarg != NULL)
					benchmark = atoi(optarg);
				else
					benchmark = 1;
				break;
			default:
				LOG(stderr, "Usage: eternity2 [--benchmark] [--gui] [--minigui]\n");
				exit(2);
		}
	
	if (gui && minigui) error("you have to choose between gui and minigui");
	
	/* if launched with /s or /S it is a windows screensaver */
	if (argc > 1 && strcasecmp("/s", argv[1]) == 0)
	{
		gui = 1;
		screensaver = 1;
	}
	#endif
	
	/* init all */
	init();
	
	/* go! */
	for (;;)
	{
		shuffle();
		fill(0, -WIDTH-HEIGHT, 0);
		if (benchmark) break;
		LOG(stderr, "Restarting...\n");
	}
	
	/* quit, will this ever happen? */
	quit();
	return 0;
}
